""":class:`Handler` and related methods"""
import typing

from pydantic import create_model, Extra, BaseConfig, BaseModel, Field
from pydantic.fields import FieldInfo
from pydantic.error_wrappers import ErrorWrapper, ValidationError
from fastapi.dependencies.utils import get_typed_signature, get_param_field
from fastapi.routing import _prepare_response_content
from fastapi.utils import create_response_field

from .models import WebSocketEventMessage

class Handler:
    """
    Class representation of a handler. It holds information about the handler, e.g. input model
    (based on :class:`pydantic.BaseModel`), :param:`event`, etc
    """
    def __init__(
        self,
        event: str,
        method: typing.Callable,
        response_model: typing.Type[BaseModel] | None = None
    ) -> None:
        self.event = event
        self.method = method

        self.bound_method: typing.Callable | None = None

        # create EventMessage model for input validation
        self.model = create_model(f'WebSocketEventMessage_{self.event}',
            type=(typing.Literal[self.event],...),
            __config__=type("Config", (BaseConfig,), {'extra': Extra.forbid})
        )

        # add all arguments (except for self) to the model
        signature = get_typed_signature(self.method)
        for param_name, param in signature.parameters.items():
            if param_name == 'self':
                continue
            field = get_param_field(param_name=param_name, param=param)
            self.model.__fields__[param_name] = field

        # create response_model if we didn't get one
        if response_model is None:
            self.response_model = create_model(
                f"Response_{event}",
                type=event,
                __config__=type("Config", (BaseConfig,), {'extra': Extra.allow})
            )
        else:
            self.response_model = response_model

        # ensure type is in there
        if 'type' not in self.response_model.__fields__:
            self.response_model.__fields__['type'] = Field(name='type',
                    type_=str,
                    required=False,
                    default=event,
                    field_info = FieldInfo(None)
            )

        self.response_field = create_response_field(
            name=f"Response_{self.event}",
            type_=self.response_model,
            required=True
        )

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Generator:
        method = self.method if self.bound_method is None else self.bound_method
        return await method(*args, **kwargs)

    async def handle(self, msg: WebSocketEventMessage) -> BaseModel | None:
        errors = []
        field = self.response_field
        data = self.model.parse_obj(msg).dict(exclude={'type'})
        response_content = _prepare_response_content(
            await self(**data),
            exclude_unset=False,
            exclude_defaults=False,
            exclude_none=False
        )

        if response_content is None:
            return None

        value, errors_ = field.validate(response_content, {}, loc=("response",))
        if isinstance(errors_, ErrorWrapper):
            errors.append(errors_)
        elif isinstance(errors_, list):
            errors.extend(errors_)
        if errors:
            raise ValidationError(errors, field.type_)
        return value

def on_event(event: str, response_model: typing.Type[BaseModel] | None = None) -> typing.Callable:
    """
    Should only be used in subclasses of :class:`WebSocketHandlingEndpoint`
    Declares a method as handler for :param:`event`

    Technical Note: Since it's impossible to get the class of an unbound function this decorator
    just sets some attributes on the function. The registration as handler happens in
    :meth:`HandlingEndpointMeta.__new__`
    """
    def decorator(func: typing.Callable) -> Handler:
        return Handler(event, func, response_model = response_model)
    return decorator



"""This module provides the :class:`WebSocketHandlingEndpoint`"""
import typing
import json
import logging
from types import MethodType

from starlette import status
from starlette.types import Receive, Scope, Send
from starlette.exceptions import HTTPException
from starlette.websockets import WebSocket, WebSocketDisconnect
from pydantic import ValidationError, create_model, BaseModel
from fastapi.encoders import jsonable_encoder

from .models import WebSocketEventMessage
from .handler import Handler

if typing.TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict

class HandlingEndpointMeta(type):
    def __new__(cls: typing.Type[type], *args: str, **kwargs: typing.Any) -> type:
        endpoint = type.__new__(cls, *args, **kwargs)

        handlers = {}

        for methodname in dir(endpoint):
            method = getattr(endpoint, methodname)

            # convert on_... methods to Handlers
            if methodname.startswith('on_') and methodname not in \
                    ['on_connect', 'on_receive', 'on_disconnect', 'on_event']:
                assert callable(method), 'handler methods have to be callable'
                method = Handler(methodname[3:], method)
                setattr(endpoint, methodname, method)

            if isinstance(method, Handler):
                assert method.event not in handlers
                handlers[method.event] = method

        setattr(endpoint, 'handlers', handlers)
        return endpoint

class WebSocketHandlingEndpoint(metaclass=HandlingEndpointMeta):
    """
    The WebSocketHandlingEndpoint is a class for the creation of a simple JSON-based WebSocket API

    This class is based on :class:`starlette.endpoints.WebSocketEndpoint`
    Incoming messages have to be based on :class:`WebSocketEventMessage`

    :meth:`dispatch` will call handlers based on the incoming :attr:`WebSocketEventMessage.type`.
    If the handler returns something it will be send to the client.

    To register a handler name it on_[type] or decorate it with :meth:`handler` and the
    :meth:`__init__` function will take care of it.


    You can override :meth:`on_connect` and :meth:`on_disconnect` to change what happens when
    clients connect or disconnect.
    """
    handlers: typing.Dict[str, Handler] = {}

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "websocket"
        self.scope = scope
        self.receive = receive
        self.send = send
        self.websocket = WebSocket(self.scope, receive=self.receive, send=self.send)

        # add all available events to our model
        self.event_message_model = create_model(
            'WebSocketEventMessage',
            type=(typing.Literal[tuple(self.handlers.keys())], ...),
            __base__=WebSocketEventMessage
        )

        # we need to tell give the handlers some bound methods
        for handler in self.handlers.values():
            # check if handler.method is on of our methods
            if handler in self.__class__.__dict__.values() \
                   and not isinstance(handler.method, (classmethod, staticmethod)):
                handler.bound_method = MethodType(handler.method, self)

    def __await__(self) -> typing.Generator:
        return self.dispatch().__await__()

    @classmethod
    def on_event(
        cls,
        event: str,
        response_model: typing.Type[BaseModel] | None = None
    ) -> typing.Callable:
        """
        Declares a method as handler for :param:`event`
        """
        def decorator(func: typing.Callable) -> Handler:
            assert event not in cls.handlers
            handler = Handler(event, func, response_model = response_model)
            cls.handlers[event] = handler
            return handler
        return decorator

    async def dispatch(self) -> None:
        """
        Handles the lifecycle of a :class:`WebSocket` connection and calls :meth:`on_connect`,
        the :meth:`handle` and :meth:`on_disconnect` repectively.

        This method will be called by starlette.

        If the client sends a JSON payload that does not conform to :class:`WebSocketEventMessage`
        or :meth:`handle` raises an :exc:`Exception` the errors will be send to the client via
        :meth:`send_exception`.
        """
        await self.on_connect()

        close_code = status.WS_1000_NORMAL_CLOSURE

        try:
            while True:
                try:
                    data = await self.websocket.receive_json()
                    msg = self.event_message_model(**data)
                    await self.handle(msg)
                except json.decoder.JSONDecodeError as exc:
                    await self.websocket.close(code=status.WS_1003_UNSUPPORTED_DATA)
                    raise RuntimeError("Malformed JSON data received.") from exc
                except ValidationError as exc:
                    await self.send_exception(exc)
                except WebSocketDisconnect as exc:
                    close_code = exc.code
                except Exception as exc:
                    raise exc

        except Exception as exc:
            close_code = status.WS_1011_INTERNAL_ERROR
            raise exc
        finally:
            await self.on_disconnect(close_code)

    async def handle(self, msg: WebSocketEventMessage) -> None:
        """Calls the handler for the incoming ``msg``"""
        assert msg.type in self.handlers, 'handle called for unknown event'
        logging.debug("Calling handler for message %s", msg)

        response = await self.handlers[msg.type].handle(msg)

        if response is not None:
            await self.send_json(response)

    async def send_exception(self, exc: Exception) -> None:
        """
        Formats the ``exc`` and sends it to the client (via :meth:`send_json`)

        Override if you don't wnat to send any Exceptions to the client or want to format them
        differently.
        """
        errors: typing.List[typing.Dict[str,typing.Any]] | typing.List[ErrorDict] # pylint: disable=used-before-assignment

        if isinstance(exc, ValidationError):
            errors = exc.errors()
        elif isinstance(exc, HTTPException):
            errors = [{
                'msg': exc.detail,
                'status_code': exc.status_code,
                'type': type(exc).__name__}]
        else:
            errors = [{'msg': str(exc), 'type': type(exc).__name__}]

        await self.send_json({'errors': errors})

    async def send_json(self, response: typing.Any) -> None:
        """
        Override to handle outgoing messages

        For example you could handle handler response differently based on their type.
        """
        return await self.websocket.send_json(jsonable_encoder(response))

    async def on_connect(self) -> None:
        """Override to handle an incoming websocket connection"""
        await self.websocket.accept()

    async def on_disconnect(self, close_code: int) -> None:
        """Override to handle a disconnecting websocket"""



""":module:`pydantic` models used in this project"""
from pydantic import BaseModel

class WebSocketEventMessage(BaseModel):
    """
    BaseModel of an incoming WebSocketEvent

    All other arguments of a handler will be added dynamically.

    Note: When validating incoming messages, type will be replaced with an :class:`typing.Literal`
    for all registered event types.

    So basically this whole class is kind of pointless.
    """
    type: str

    class Config:
        use_enum_values = True
        extra = 'allow'
